#import <QuartzCore/QuartzCore.h>
#import <CoreText/CoreText.h>
#import "ABCTextView.h"

#define EMPTY @""

NSString *const kABCTextViewTypeText = @"text";
NSString *const kABCTextViewTypeBackground = @"background";
NSString *const kABCTextViewTypeComment = @"comment";
NSString *const kABCTextViewTypeDocumentationComment = @"documentation_comment";
NSString *const kABCTextViewTypeDocumentationCommentKeyword = @"documentation_comment_keyword";
NSString *const kABCTextViewTypeString = @"string";
NSString *const kABCTextViewTypeCharacter = @"character";
NSString *const kABCTextViewTypeNumber = @"number";
NSString *const kABCTextViewTypeKeyword = @"keyword";
NSString *const kABCTextViewTypePreprocessor = @"preprocessor";
NSString *const kABCTextViewTypeURL = @"url";
NSString *const kABCTextViewTypeAttribute = @"attribute";
NSString *const kABCTextViewTypeOther = @"other";

@interface ABCTextView() {
    id internalDelegate;
}
- (NSAttributedString*)highlightText:(NSAttributedString*)stringIn;
- (NSRange)visibleRangeOfTextView:(UITextView*)textView;
+ (NSDictionary*)defaultDefinition;
@end

@interface ABCTextViewDelegate : NSObject<UITextViewDelegate>
@end
@implementation ABCTextViewDelegate
//Update the syntax highlighting if the text gets changed or the scrollview gets updated
- (void)textViewDidChange:(UITextView *)textView
{
  [textView setNeedsDisplay];
}

- (void)scrollViewDidScroll:(UIScrollView *)scrollView
{
	[scrollView setNeedsDisplay];
}

- (BOOL)textView:(UITextView *)textView shouldChangeTextInRange:(NSRange)range replacementText:(NSString *)text
{
  //Only update the text if the text changed
	NSString* newText = [text stringByReplacingOccurrencesOfString:@"\t" withString:@"    "];
	if(![newText isEqualToString:text]) {
		textView.text = [textView.text stringByReplacingCharactersInRange:range withString:newText];
		return NO;
	}
	return YES;
}
@end

static CGFloat MARGIN = 8;
static NSMutableDictionary* highlightThemes;

@implementation ABCTextView
@synthesize highlightColor;
@synthesize highlightDefinition;

-(void)setHighlightColor:(NSDictionary*)newHighlightColor {
  if(highlightColor!=newHighlightColor) {
      highlightColor = newHighlightColor;
      [self setNeedsLayout];
  }
}

-(void)setHighlightDefinition:(NSDictionary*)newHighlightDefinition {
  if(highlightDefinition!=newHighlightDefinition) {
      highlightDefinition = newHighlightDefinition;
      [self setNeedsLayout];
  }
}

-(id)init
{
  self = [super init];
  if (self) {
  }
  return self;
}

-(id)initWithCoder:(NSCoder*)decoder
{
  self = [super initWithCoder:decoder];
  if (self) {
    self.textColor = [UIColor clearColor];
//    self.delegate = (internalDelegate=[[ABCTextViewDelegate alloc] init]);
  }
  return self;
}

-(id)initWithFrame:(CGRect)frame
{
  self = [super initWithFrame:frame];
  if (self) {
    self.textColor = [UIColor clearColor];
//    self.delegate = (internalDelegate=[[ABCTextViewDelegate alloc] init]);
  }
  return self;
}

-(void)drawRect:(CGRect)rect
{
  if(self.text.length<=0) {
      self.text = EMPTY;
      return;
  }

  //Prepare View for drawing
  CGContextRef context = UIGraphicsGetCurrentContext();
  CGContextSetTextMatrix(context,CGAffineTransformIdentity);
  CGContextTranslateCTM(context,0,([self bounds]).size.height);
  CGContextScaleCTM(context,1.0,-1.0);

  //Get the view frame size
  CGSize size = self.frame.size;
  
  //Determine default text color
  UIColor* textColor = nil;
  if(!self.highlightColor||!(textColor=[self.highlightColor objectForKey:kABCTextViewTypeText])) {
    if([self.textColor isEqual:[UIColor clearColor]]) {
      if(!(textColor=[[ABCTextView highlightTheme:kABCTextViewThemeDefault] objectForKey:kABCTextViewTypeText]))
         textColor = [UIColor blackColor];
    }
    else
      textColor = self.textColor;
  }
  
  //Set line height, font, color and break mode
  CGFloat minimumLineHeight = [self.text sizeWithFont:self.font].height,maximumLineHeight = minimumLineHeight;
  CTFontRef font = CTFontCreateWithName((CFStringRef)self.font.fontName,self.font.pointSize,NULL);
  CTLineBreakMode lineBreakMode = kCTLineBreakByWordWrapping;
  
  //Apply paragraph settings
  CTParagraphStyleRef style = CTParagraphStyleCreate((CTParagraphStyleSetting[3]){
      {kCTParagraphStyleSpecifierMinimumLineHeight,sizeof(minimumLineHeight),&minimumLineHeight},
      {kCTParagraphStyleSpecifierMaximumLineHeight,sizeof(maximumLineHeight),&maximumLineHeight},
      {kCTParagraphStyleSpecifierLineBreakMode,sizeof(CTLineBreakMode),&lineBreakMode}
  },3);
  NSDictionary* attributes = [NSDictionary dictionaryWithObjectsAndKeys:(id)font,(NSString*)kCTFontAttributeName,(id)textColor.CGColor,(NSString*)kCTForegroundColorAttributeName,(id)style,(NSString*)kCTParagraphStyleAttributeName,nil];
  CFRelease(font);
  CFRelease(style);

  //Create path to work with a frame with applied margins
  CGMutablePathRef path = CGPathCreateMutable();
  CGPathAddRect(path,NULL,CGRectMake(MARGIN+0.0,(-self.contentOffset.y+0),(size.width-2*MARGIN),(size.height+self.contentOffset.y-MARGIN)));
      
  //Create attributed string, with applied syntax highlighting
  NSAttributedString *attributedString = [self highlightText:[[[NSAttributedString alloc] initWithString:self.text attributes:attributes] autorelease]];

  //Draw the frame
  CTFramesetterRef framesetter = CTFramesetterCreateWithAttributedString((CFAttributedStringRef)attributedString);
  CTFrameRef frame = CTFramesetterCreateFrame(framesetter, CFRangeMake(0,[attributedString length]),path,NULL);
   CTFrameDraw(frame,context);

  CFRelease(framesetter);
  CFRelease(frame);
  CGPathRelease(path);
}

-(NSRange)visibleRangeOfTextView:(UITextView *)textView
{
  CGRect bounds = textView.bounds;
  //Get start and end bouns for text position
  UITextPosition *start = [textView characterRangeAtPoint:bounds.origin].start,*end = [textView characterRangeAtPoint:CGPointMake(CGRectGetMaxX(bounds),CGRectGetMaxY(bounds))].end;
  //Make a range out of it and return
  return NSMakeRange([textView offsetFromPosition:textView.beginningOfDocument toPosition:start],[textView offsetFromPosition:start toPosition:end]);
}

-(NSAttributedString*)highlightText:(NSAttributedString*)attributedString
{
  //Create a mutable attribute string to set the highlighting
  NSString* string = attributedString.string; NSRange range = NSMakeRange(0,[string length]);
  NSMutableAttributedString* coloredString = [[NSMutableAttributedString alloc] initWithAttributedString:attributedString];
  
  //Define the definition to use
  NSDictionary* definition = nil;
  if(!(definition=self.highlightDefinition))
      definition = [ABCTextView defaultDefinition];
  
  //For each definition entry apply the highlighting to matched ranges
  for(NSString* key in definition) {
      NSString* expression = [definition objectForKey:key];
      if(!expression||[expression length]<=0) continue;
      NSArray* matches = [[NSRegularExpression regularExpressionWithPattern:expression options:NSRegularExpressionAnchorsMatchLines error:nil] matchesInString:string options:0 range:range];
      for(NSTextCheckingResult* match in matches) {
          UIColor* textColor = nil;
          //Get the text color, if it is a custom key and no color was defined, choose black
          if(!self.highlightColor||!(textColor=([self.highlightColor objectForKey:key])))
              if(!(textColor=[[ABCTextView highlightTheme:kABCTextViewThemeDefault] objectForKey:key]))
                  textColor = [UIColor blackColor];
          [coloredString addAttribute:(NSString*)kCTForegroundColorAttributeName value:(id)textColor.CGColor range:[match rangeAtIndex:0]];                            
      }
  }
  
  return [[[coloredString autorelease] copy] autorelease];
}

-(void)setHighlightTheme:(ABCTextViewTheme)theme
{
  self.highlightColor = [ABCTextView highlightTheme:theme];
  
  //Set font, text color and background color back to default
  self.textColor = [UIColor clearColor];
  UIColor* backgroundColor = [self.highlightColor objectForKey:kABCTextViewTypeBackground];
  if (backgroundColor)
    self.backgroundColor = backgroundColor;
  else
    self.backgroundColor = [UIColor whiteColor];
  self.font = [UIFont systemFontOfSize:10.0];
}

+(NSDictionary*)highlightTheme:(ABCTextViewTheme)theme
{
    //Check if the highlight theme has already been defined
    NSDictionary* themeColor = nil;
    if(!highlightThemes)
      highlightThemes = [[NSMutableDictionary dictionary] retain];
    if ((themeColor = [highlightThemes objectForKey:[NSNumber numberWithInt:theme]]))
        return themeColor;
    
    //If not define the theme and return it
    switch(theme) {
    case kABCTextViewThemeDefault:
        themeColor = [NSDictionary dictionaryWithObjectsAndKeys:
                [UIColor colorWithRed:0.0/255 green:0.0/255 blue:0.0/255 alpha:1],kABCTextViewTypeText,
                [UIColor colorWithRed:255.0/255 green:255.0/255 blue:255.0/255 alpha:1],kABCTextViewTypeBackground,
                [UIColor colorWithRed:0.0/255 green:131.0/255 blue:39.0/255 alpha:1],kABCTextViewTypeComment,
                [UIColor colorWithRed:0.0/255 green:131.0/255 blue:39.0/255 alpha:1],kABCTextViewTypeDocumentationComment,
                [UIColor colorWithRed:0.0/255 green:76.0/255 blue:29.0/255 alpha:1],kABCTextViewTypeDocumentationCommentKeyword,
                [UIColor colorWithRed:211.0/255 green:45.0/255 blue:38.0/255 alpha:1],kABCTextViewTypeString,
                [UIColor colorWithRed:40.0/255 green:52.0/255 blue:206.0/255 alpha:1],kABCTextViewTypeCharacter,
                [UIColor colorWithRed:255.0/255 green:52.0/255 blue:52.0/255 alpha:1],kABCTextViewTypeNumber,
                [UIColor colorWithRed:188.0/255 green:49.0/255 blue:156.0/255 alpha:1],kABCTextViewTypeKeyword,
                [UIColor colorWithRed:120.0/255 green:72.0/255 blue:48.0/255 alpha:1],kABCTextViewTypePreprocessor,
                [UIColor colorWithRed:21.0/255 green:67.0/255 blue:244.0/255 alpha:1],kABCTextViewTypeURL,
                [UIColor colorWithRed:150.0/255 green:125.0/255 blue:65.0/255 alpha:1],kABCTextViewTypeAttribute,
                [UIColor colorWithRed:113.0/255 green:65.0/255 blue:163.0/255 alpha:1],kABCTextViewTypeOther,nil];
        break;
    }
    if(themeColor) {
      [highlightThemes setObject:themeColor forKey:[NSNumber numberWithInt:theme]];
      return themeColor;
    }
    else
      return nil;
}

+(NSDictionary*)defaultDefinition {
    //It is recommended to use an ordered dictionary, because the highlighting will take place in the same order the dictionary enumerator returns the definitions
    NSMutableDictionary* definition = [NSMutableDictionary dictionary];
    [definition setObject:@"^[A-Za-z]:.*?$" forKey:kABCTextViewTypeKeyword];
    [definition setObject:@"\\[[A-Z]:.*?\\]" forKey:kABCTextViewTypeDocumentationCommentKeyword];
//    [definition setObject:@"((https?|mailto|ftp|file)://([-\\w\\.]+)+(:\\d+)?(/([\\w/_\\.]*(\\?\\S+)?)?)?)" forKey:kABCTextViewTypeURL];
    [definition setObject:@"(\"(?:[^\"\\\\]|\\\\.)*\")" forKey:kABCTextViewTypeString];
    [definition setObject:@"%(.*?)[\r\n]" forKey:kABCTextViewTypeComment];
    [definition setObject:@"(\\|:|:\\||:\\]|\\[:|::|\\|]|\\[\\||\\|[1-9]|\\[[1-9]|\\|)" forKey:kABCTextViewTypeNumber];
    return definition;
}
@end
