#import "TunePlayer.h"
#import "AppDelegate.h"
#import "fluid_midi.h"

static NSString *soundfonts[] = {
  @"GeneralUser GS FluidSynth v1.44.sf2",
};

@implementation Voice
- (id)initWithPreset:(fluid_preset_t)preset
{
  self = [super init];
  if ( self ) {
    _preset = preset;
  }
  return self;
}

- (NSString*)name
{
  return [NSString stringWithUTF8String:_preset.get_name(&_preset)];
}

- (int)bank
{
  return _preset.get_banknum(&_preset);
}

- (int)number
{
  return _preset.get_num(&_preset);
}
@end

@implementation TunePlayer

@synthesize filePath, transpose;

static void player_notify_callback( fluid_player_t* player, void *data )
{
  TunePlayer* tunePlayer = (TunePlayer*)data;
  [tunePlayer performSelectorOnMainThread:@selector(playerFinished:) withObject:tunePlayer waitUntilDone:NO];
}

+ (TunePlayer *)tunePlayer;
{
  static TunePlayer *player;
  if ( !player )
    player = [[TunePlayer alloc] init];
  return player;
}

#if TUNEBOOK_AUDIO_TOOLBOX
-(OSStatus) loadFromDLSOrSoundFont: (NSURL *)bankURL withPatch: (int)presetNumber
{
    OSStatus result = noErr;

    AUSamplerInstrumentData instdata;
    instdata.fileURL  = (CFURLRef) bankURL;
    instdata.instrumentType = kInstrumentType_DLSPreset;
    instdata.bankMSB  = kAUSampler_DefaultMelodicBankMSB;
    instdata.bankLSB  = kAUSampler_DefaultBankLSB;
    instdata.presetID = (UInt8) presetNumber;

    // set the kAUSamplerProperty_LoadPresetFromBank property
    result = AudioUnitSetProperty(samplerUnit,
                                  kAUSamplerProperty_LoadInstrument,
                                  kAudioUnitScope_Global,
                                  0,
                                  &instdata,
                                  sizeof(instdata));

    // check for errors
    NSCAssert (result == noErr,
               @"Unable to set the preset property on the Sampler. Error code:%d '%.4s'",
               (int) result,
               (const char *)&result);

    return result;
}
#endif

- (id)init
{
  self = [super init];
  if ( self ) {
    AppDelegate *appDelegate = (AppDelegate *)[[UIApplication sharedApplication] delegate];
    speed = appDelegate.playerSpeed;
    voice = appDelegate.playerVoice;
    volume = appDelegate.playerVolume;
    settings = NULL;
    synth = NULL;
    adriver = NULL;
    isPlaying = NO;
    _voices = [[NSMutableArray alloc] init];

    if ( !settings ) {
#if TUNEBOOK_AUDIO_TOOLBOX
      OSStatus status = NewAUGraph(&graph);
      AUNode samplerNode;
      AudioComponentDescription samplerDescription = {
        kAudioUnitType_MusicDevice,
        kAudioUnitSubType_Sampler,
        kAudioUnitManufacturer_Apple,
        0, 0
      };
      status = AUGraphAddNode(graph, &samplerDescription, &samplerNode);

      AUNode ioNode;
      AudioComponentDescription ioDescription = {
        kAudioUnitType_Output,
        kAudioUnitSubType_RemoteIO,
        kAudioUnitManufacturer_Apple,
        0, 0
      };
      status = AUGraphAddNode(graph, &ioDescription, &ioNode);

      status = AUGraphOpen(graph);
      status = AUGraphConnectNodeInput(graph, samplerNode, 0, ioNode, 0);
      status = AUGraphInitialize(graph);
      status = AUGraphStart(graph);

      NSString* sfFile = [[[NSBundle mainBundle] resourcePath] stringByAppendingPathComponent:soundfonts[0]];
      NSURL* sfURL = [NSURL URLWithString:sfFile];
      status = [self loadFromDLSOrSoundFont:sfURL withPatch:0];
#else
#endif
      settings = new_fluid_settings();

      fluid_settings_setstr(settings, "audio.driver", "coreaudio");
      fluid_settings_setstr(settings, "audio.sample-format", "16bits");
      fluid_settings_setnum(settings, "synth.sample-rate", 8000);

      synth = new_fluid_synth(settings);
//      adriver = new_fluid_audio_driver(settings, synth);

      NSString *sfFile = [[[NSBundle mainBundle] resourcePath] stringByAppendingPathComponent:soundfonts[0]];
      sfont_id[0] = fluid_synth_sfload(synth, [sfFile UTF8String], 0);
      for ( int i = 1; i < [self numVoices]; i++ )
        sfont_id[i] = sfont_id[0];

      // Iterate through the presets of the font
      fluid_sfont_t* sfont;
      fluid_preset_t preset;
      int offset;

      sfont = fluid_synth_get_sfont_by_id(synth, sfont_id[0]);
      offset = fluid_synth_get_bank_offset(synth, sfont_id[0]);

      sfont->iteration_start(sfont);
      while (sfont->iteration_next(sfont, &preset)) {
        if ( preset.get_banknum(&preset) == 0 ) {
          [_voices addObject:[[Voice alloc] initWithPreset:preset]];
        }
      }
    }
  }
  return self;
}

- (void)dealloc
{
#if TUNEBOOK_AUDIO_TOOLBOX
  DisposeAUGraph(graph);
#else
#endif
  if ( adriver )
    delete_fluid_audio_driver( adriver );
  if ( synth )
    delete_fluid_synth( synth );
  if ( settings )
    delete_fluid_settings( settings );
  [super dealloc];
}

- (int)numVoices
{
  return (int)[_voices count];
}

- (NSString *)voiceName:(int)index
{
  if ( index >= 0 && index < [_voices count] ) {
    Voice* v = [_voices objectAtIndex:index];
    return [v name];
  }
  return @"Unknown";
}

- (NSArray*)voices
{
  return _voices;
}

- (void)setSpeed:(float)s
{
  AppDelegate *appDelegate = (AppDelegate *)[[UIApplication sharedApplication] delegate];
  appDelegate.playerSpeed = speed = s;
  if ( player )
    fluid_player_set_speed_multiplier( player, pow( 2.0, speed ) );
}

- (float)speed
{
  return speed;
}

- (float)time
{
  if ( player ) {
    if ( fluid_player_count_tracks(player) > 0 ) {
      fluid_track_t* track = fluid_player_get_track(player, 0);
      return (float)player->cur_ticks / (float)fluid_track_get_duration(track);
    }
  }
  return 0;
}

- (void)setTime:(float)time
{
  if ( player ) {
    if ( fluid_player_count_tracks(player) > 0 ) {
      fluid_track_t* track = fluid_player_get_track(player, 0);
      fluid_player_seek( player, (int)(time * fluid_track_get_duration(track)) );
    }
  }
}

- (void)setVoice:(int)index
{
  AppDelegate *appDelegate = (AppDelegate *)[[UIApplication sharedApplication] delegate];
  appDelegate.playerVoice = voice = index;
  if ( synth ) {
    for ( int channel = 0; channel < 13; channel++ ) {
      Voice* v = [_voices objectAtIndex:index];
      fluid_synth_program_select(synth, channel, sfont_id[0], [v bank], [v number]);
    }
  }
}

- (int)voice
{
  return voice;
}

- (void)setTrebleVoice:(int)index
{
  AppDelegate *appDelegate = (AppDelegate *)[[UIApplication sharedApplication] delegate];
  appDelegate.playerVoiceTrebleAccompaniment = trebleVoice = index;
  if ( synth ) {
    Voice* v = [_voices objectAtIndex:index];
    fluid_synth_program_select(synth, 14, sfont_id[0], [v bank], [v number]);
  }
}

- (void)setBassVoice:(int)index
{
  AppDelegate *appDelegate = (AppDelegate *)[[UIApplication sharedApplication] delegate];
  appDelegate.playerVoiceBassAccompaniment = bassVoice = index;
  if ( synth ) {
    Voice* v = [_voices objectAtIndex:index];
    fluid_synth_program_select(synth, 15, sfont_id[0], [v bank], [v number]);
  }
}

- (void)setVolume:(float)v
{
  AppDelegate *appDelegate = (AppDelegate *)[[UIApplication sharedApplication] delegate];
  appDelegate.playerVolume = volume = v;
  if ( settings )
    fluid_settings_setnum( settings, "synth.gain", volume * 0.25 );
}

- (float)volume
{
  return volume;
}

- (void)play
{
  [self playFromTime:0.0f];
}

- (void)playFromTime:(float)time
{
  [self stop];
  if ( !adriver )
    adriver = new_fluid_audio_driver(settings, synth);

  AppDelegate *appDelegate = (AppDelegate *)[[UIApplication sharedApplication] delegate];
  player = new_fluid_player( synth );
  fluid_player_set_notify_callback( player, player_notify_callback, self );
  [self setVoice:appDelegate.playerVoice];
  [self setTrebleVoice:appDelegate.playerVoiceTrebleAccompaniment];
  [self setBassVoice:appDelegate.playerVoiceBassAccompaniment];
  fluid_player_add( player, (char *)[filePath UTF8String] );
  fluid_player_set_speed_multiplier( player, pow( 2.0, speed ) );
  fluid_settings_setnum( settings, "synth.gain", volume * 0.25 );
  fluid_player_set_loop( player, appDelegate.loopPlayback );
  fluid_player_initial_seek( player, time );
  fluid_player_play( player );
  isPlaying = YES;
  isPaused = NO;
  [[NSNotificationCenter defaultCenter] postNotificationName:@"startedPlaying" object:self];
}

- (void)stop
{
  if ( player ) {
    fluid_player_stop(player);
    fluid_player_join(player);
    delete_fluid_player(player);
    fluid_synth_system_reset( synth );
    player = NULL;
  }
  if ( isPlaying )
    [[NSNotificationCenter defaultCenter] postNotificationName:@"finishedPlaying" object:self];
  isPlaying = NO;
  isPaused = NO;
}

- (void)reset
{
  if ( player ) {
    fluid_player_stop(player);
    fluid_player_join(player);
    delete_fluid_player(player);
    fluid_synth_system_reset( synth );
    player = NULL;
  }
  if ( isPlaying )
    [[NSNotificationCenter defaultCenter] postNotificationName:@"finishedPlaying" object:self];
  isPlaying = NO;
  isPaused = NO;
  pauseTime = 0;
}

- (void)pause
{
  //FIXME-testing
  if ( isPlaying ) {
//    fluid_player_stop( player );
    pauseTime = player->cur_ticks;
    [self stop];
    isPaused = YES;
  }
}

- (BOOL)playing
{
  return player != NULL;
}

- (BOOL)paused
{
  return isPaused;
}

- (void)playerFinished:(id)sender
{
  [self stop];
}

@end
