/*
 *  abc2ps: a program to typeset tunes written in abc format using PostScript
 *  Copyright (C) 1996-2007  Michael Methfessel
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  The author can be contacted as follows:
 *
 *  Michael Methfessel
 *  methfessel at ihp-ffo dot de
 *  IHP
 *  Im Technologiepark 25
 *  D-15236 Frankfurt (Oder), Germany
 */

/* Main program abc2ps.c */

#include <string.h>

#include "abc2ps.h"
#include "subs.h"
#include "format.h"
#include "util.h"
#include "parse.h"
#include "music.h"
#include "playback.h"


/* ----- definitions of global variables ----- */

int db = DEBUG_LV;                  /* debug control */
int maxSyms, maxVc;                /* for malloc */
InformationFields info, default_info; /* information fields */
char lvoiceid[233];               /* string from last V: line */
int  nvoice, mvoice;               /* number of voices defd, nonempty */
int  ivc;                         /* current voice */
int  ivc0;                        /* top nonempty voice */
int ixpfree;                      /* first free element in xp array */
TimeSignature default_meter;    /* data to specify the meter */
KeySignature default_key;        /* data to specify the key */

char gTuneParts[STRL];
char gCurrentPart;

/* things to alloc: */
Symbol* sym;              /* symbol list */
Symbol** symv;            /* symbols for voices */
XPosition* xp;               /* shared horizontal positions */
Voice* voice;            /* characteristics of a voice */
Symbol** sym_st;           /* symbols a staff start */
int* nsym_st;

Symbol zsym;            /* symbol containing zeros */

Format sfmt;                    /* format after initialization */
Format dfmt;                    /* format at start of tune */
Format cfmt;                    /* current format for output */

char fontnames[50][STRLFMT];    /* list of needed fonts */
int  nfontnames;

char buf[BUFFSZ];              /* output buffer.. should hold one tune */
int nbuf;                      /* number of bytes buffered */
float bposy;                   /* current position in buffered data */
int   use_buffer;              /* 1 if lines are being accumulated */

char text [NTEXT][STRL];       /* pool for history, words, etc. lines */
int text_type[NTEXT];          /* type of each text line */
int ntext;                     /* number of text lines */
int linenum;                   /* current line number in input file */
int tunenum;                   /* number of current tune */
int numtitle;                  /* how many titles were read */
int mline;                     /* number music lines in current tune */
int nsym;                      /* number of symbols in line */
int nsym0;                     /* nsym at start of parsing a line */
int pagenum;                   /* current page in output file */
int writenum;                  /* calls to write_buffer for each one tune */
int xrefnum;                   /* xref number of current tune */
int do_meter, do_indent;       /* how to start next block */

int bagpipe;                     /* switch for HP mode */
int within_tune, within_block;   /* where we are in the file */
int do_this_tune;                /* are we typesetting the current one ? */
float posx, posy;                 /* overall scale, position on page */
int barinit;                     /* carryover bar number between parts */

int num_ending;                  /* number of endings to draw */
int mes1, mes2;                   /* to count measures in an ending */

int overfull;                    /* flag if staff overfull */
int do_words;                    /* flag if staff has words under it */

int vb, verbose;                 /* verbosity, global and within tune */
int in_page = 0;

/* switches modified by flags: */
char znote[201];                   /* annotation at top of page */
int select_all;                    /* select all tunes ? */
int respect_case;                  /* does case matter in matching ? */
int search_field0;                 /* default search field */
int pretty;                        /* for pretty but sprawling layout */

float alfa_last, beta_last;         /* for last short short line.. */

char outf[STRL1];                /* output file name */
char styf[STRL1];                /* layout style file name */
char styd[STRL1];                /* layout style directory */

int  file_open;                  /* for output file */
int  file_initialized;           /* for output file */
FILE* fout, * ftext;       /* for output file */

#if !ABC_LIBRARY
char sel_str[MAXINF][STRL1];     /* list of selector strings */
int  s_field[MAXINF];            /* type of selection for each file */
int  psel[MAXINF];               /* pointers from files to selectors */
#endif

char lout[BSIZE];

int utf8charset;

ABCMusicSettings musicSettings;
//ABCParser parser;

GraphicsContext gcx;
#if ABC_INPUT_STRING
const char* abcInputString;
#endif

#if ABC_LIBRARY
enum {
  kABCPaginate = 0x01,
  kABCOutputMIDI = 0x02,
  kABCOutputGuitarChords = 0x04,
  kABCOutputABC = 0x08,
};

int ABCTransformABC( const char* abc, const char* midiFile, const char* spec )
{
  in_page = 0;
  ivc = 0;
  utf8charset = 1;
  resetRandom();

  char pat[30][STRL1];

  init_ops (1);

  respect_case = 0;
  strcpy (znote, "");
  musicSettings.global_halftones = 0;
  musicSettings.do_mode = DO_OUTPUT;
  maxVc = 8;
  maxSyms = 2400;

  alfa_last = 0;
  beta_last = 0;
  ixpfree = 0;
  linenum = 0;
  mline = 0;
  nsym = 0;
  posy = 0;
  nvoice = 0;
  mvoice = 0;
  ivc = 0;
  ivc0 = 0;
  gTuneParts[0] = 0;
  gCurrentPart = 'A';

#if TARGET_OS_IPHONE
  //An attempt to fix memory fragmentation which eventually causes a crash
  static int firstTime = 1;
  if ( firstTime )
    alloc_structs ();
  firstTime = 0;
#else
  alloc_structs ();
#endif

  nfontnames = 0;

  abcInputString = abc;

  /* ----- initialize ----- */
  zero_sym ();
  pagenum = 0;
  tunenum = 0;
  verbose = 0;
  file_open = file_initialized = 0;
  bposy = 0;
  posx = cfmt.leftmargin;
  posy = cfmt.pageheight - cfmt.topmargin;

  musicSettings.write_text = FALSE;

  musicSettings.write_text = 1;
  musicSettings.write_music = 0;
  musicSettings.make_audio = 0;

  strcpy (musicSettings.global_transpose, spec);
  musicSettings.ntrp = 4;
  for (int i = 0; i < musicSettings.ntrp; i++)
    strcpy (musicSettings.transpose[i], spec);

  select_all = TRUE;
  dfmt = sfmt;

  pagenum = 0;
  tunenum = 0;
  file_open = 1;
  file_initialized = 0;

  ftext = fopen( midiFile, "w" );
  if ( ftext != 0 ) {
    process_file (0, fout, "", 0, pat, select_all, 0);
    fclose( ftext );
  }

  //  free_structs();
  return 0;
}

int ABCTypesetTune( const char* abc, const char* format, const char* midiFile, int flags )
{
  in_page = 0;
  ivc = 0;
  utf8charset = 1;
  resetRandom();

  char pat[30][STRL1];

  init_ops (1);

  respect_case = 0;
  strcpy (znote, "");
  strcpy (musicSettings.global_transpose, "");
  musicSettings.global_halftones = 0;
  musicSettings.do_mode = DO_OUTPUT;
  maxVc = 8;
  maxSyms = 2400;

  alfa_last = 0;
  beta_last = 0;
  ixpfree = 0;
  linenum = 0;
  mline = 0;
  nsym = 0;
  posy = 0;
  nvoice = 0;
  mvoice = 0;
  ivc = 0;
  ivc0 = 0;
  gTuneParts[0] = 0;
  gCurrentPart = 'A';

  #if TARGET_OS_IPHONE
  //An attempt to fix memory fragmentation which eventually causes a crash
  static int firstTime = 1;
  if ( firstTime )
    alloc_structs ();
  firstTime = 0;
  #else
  alloc_structs ();
  #endif

  nfontnames = 0;
  if ( format ) {
    abcInputString = format;
    char line[BSIZE];
    int i, end;
    strcpy (line, "");
    abcgetline (line, BSIZE, 0);
    for (i = 0; i < 200; i++) {
      end = interpret_format_line (line, &cfmt);
      if (end == 1)
        return 1;
      strcpy (line, "");
      if (!abcgetline(line, BSIZE, 0))
        return 1 ;
    }
    ops_into_fmt (&cfmt);

    make_font_list (&cfmt);
    sfmt = cfmt;
    dfmt = cfmt;
  }
  else if (!set_page_format())
    return -1;

  abcInputString = abc;

  /* ----- initialize ----- */
  zero_sym ();
  pagenum = 0;
  tunenum = 0;
  verbose = 0;
  file_open = file_initialized = 0;
  bposy = 0;
  posx = cfmt.leftmargin;
  posy = cfmt.pageheight - cfmt.topmargin;

#if 0//ABC_PLAYBACK_PARTS//FIXME-testing
vb = 10;//FIXME
verbose = 17;
//printf("%ld\n", sizeof(Symbol));
#endif

#if 1 //MIDI_OUTPUT
  if ( midiFile && (flags & kABCOutputMIDI) != 0 ) {
    musicSettings.make_audio = 1;
    musicSettings.write_music = 0;
    musicSettings.midi_format = 1;
    musicSettings.do_acc = (flags & kABCOutputGuitarChords) != 0;
    musicSettings.au_nomusic = 0;
  }
  else if ( midiFile && (flags & kABCOutputABC) != 0 ) {
    musicSettings.write_text = 1;
    musicSettings.write_music = 0;
    musicSettings.make_audio = 0;
    ftext = stdout;
  }
  else {
    musicSettings.write_text = FALSE;
    musicSettings.make_audio = 0;
    musicSettings.write_music = 1;
  }
  if ((musicSettings.do_mode == DO_OUTPUT) && musicSettings.make_audio)
    open_audio_file(midiFile);
#endif

#if ABC_USE_CAIRO
  int outputWidth;
  int outputHeight;
  cairo_surface_t* surface;
  int scale = 1;
  if ( !musicSettings.outputFormat )
    musicSettings.outputFormat = "png";
  if ( strcmp( musicSettings.outputFormat, "png" ) == 0 )
    scale = 2;
  outputWidth = scale * (cfmt.staffwidth + 2 * cfmt.leftmargin);
  outputHeight = scale * cfmt.pageheight;
  if ( strcmp( musicSettings.outputFormat, "png" ) == 0 )
    surface = cairo_image_surface_create( CAIRO_FORMAT_RGB24, outputWidth, outputHeight );
  else if ( strcmp( musicSettings.outputFormat, "pdf" ) == 0 )
    surface = cairo_pdf_surface_create( "Out.pdf", outputWidth, outputHeight);
  else if ( strcmp( musicSettings.outputFormat, "ps" ) == 0 )
    surface = cairo_ps_surface_create( "Out.ps", outputWidth, outputHeight);
  else if ( strcmp( musicSettings.outputFormat, "svg" ) == 0 )
    surface = cairo_svg_surface_create( "Out.svg", outputWidth, outputHeight);
  gcx = cairo_create( surface );
  cairo_set_source_rgb(gcx, 1, 1, 1);
  cairo_rectangle(gcx, 0, 0, outputWidth, outputHeight);
  cairo_fill (gcx);
  cairo_scale( gcx, scale, scale );
  cairo_scale( gcx, 1, -1 );
  cairo_set_source_rgb(gcx, 0, 0, 0);
#endif

  select_all = TRUE;
  dfmt = sfmt;

  pagenum = 0;
  tunenum = 0;
  file_open = 1;
  file_initialized = 0;

  if ( flags == 0 ) {
    dfmt.pageheight = 10000;
    cfmt.pageheight = 10000;
  }
  process_file (0, fout, "", 0, pat, select_all, 0);
  //FIXME-PM
  extern int gTuneHeight;
  gTuneHeight = dfmt.pageheight - posy;

  close_output_file ();

  if ((musicSettings.do_mode == DO_OUTPUT) && musicSettings.make_audio)
    close_audio_file ();

  #if ABC_USE_CAIRO
  if ( strcmp( musicSettings.outputFormat, "png" ) == 0 )
    cairo_surface_write_to_png( surface, "Out.png");
  cairo_surface_finish( surface );
  cairo_surface_destroy( surface );
  #endif
  //  free_structs();
  return 0;
}
#endif
