#ifndef _abc2psH
#define _abc2psH

/*
 *  This file is part of abc2ps,
 *  Copyright (C) 1996-2007  Michael Methfessel
 *  See file abc2ps.c for details.
 */

/*
 * This version of abc2ps has been extensively hacked to run on the iPhone. The main changes are the ability to
 * be built as a library instead of an application, and replacement of all the PostScript output by calls to
 * graphics functions. I tried to leave as much of the original code in place as possible to facilitate future
 * upgrades to abc2ps. However, the code is very hard to maintain and the temptation to convert the whole thing
 * to C++ and structure it well is very strong. The conversion to using Cairo as the rendering library was done
 * over three days due to time constraints, and there is much I would improve given more time.
 *
 * Jerry Huxtable
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

// Configuration:
//  TARGET_OS_IPHONE:
//    We're compiling for the iPhone/iPad
//  ABC_USE_GRAPHICS_LIB:
//    We should use the graphics calls rather than emitting PostScript. Eventually, the old PostScript stuff will be moved
//    into its own GraphicsMethods.
//  ABC_USE_CAIRO:
//    We use Cairo as the graphics library
//  ABC_LIBRARY:
//    We're building a library rather than an an application. This renames main(), turns off
//    a lot of debugging output, frees memory at the end and stops exit() being called on error.
//  ABC_INPUT_STRING:
//    A hack for the iPhone. Read the ABC input from a string instead of a file.

#if TARGET_OS_IPHONE || TARGET_OS_MAC
  #define ABC_USE_GRAPHICS_LIB 1
  #define ABC_USE_CAIRO 0
  #define ABC_LIBRARY 1
  #define ABC_INPUT_STRING 1
#else
  #ifndef ABC_USE_GRAPHICS_LIB
    #define ABC_USE_GRAPHICS_LIB 1
  #endif
  #ifndef ABC_USE_CAIRO
    #define ABC_USE_CAIRO 1
  #endif
  #ifndef ABC_LIBRARY
    #define ABC_LIBRARY 0
  #endif
  #ifndef ABC_INPUT_STRING
    #define ABC_INPUT_STRING 0
  #endif
#endif

#define ABC_PLAYBACK_PARTS 1 // Experimental support for P: in playback. Remove when it works.

#if ABC_USE_GRAPHICS_LIB
  #if ABC_USE_CAIRO
    #include "cairo.h"
    #include "cairo-ps.h"
    #include "cairo-pdf.h"
    #include "cairo-svg.h"

typedef cairo_t* GraphicsContext;
  #else
    #if TARGET_OS_IPHONE
      #include <CoreGraphics/CoreGraphics.h>
    #else
      #include <ApplicationServices/ApplicationServices.h>
    #endif

typedef CGContextRef GraphicsContext;
  #endif

enum {
  kGraphicsAlignLeft = -1,
  kGraphicsAlignCenter = 0,
  kGraphicsAlignRight = 1,
};

enum {
  kSerifFont = 0,
  kSansSerifFont = 1,
};

typedef struct
{
  void (* beginPage)( GraphicsContext cx );
  void (* endPage)( GraphicsContext cx );

  void (* moveTo)( GraphicsContext cx, float x, float y );
  void (* lineTo)( GraphicsContext cx, float x, float y );
  void (* curveTo)( GraphicsContext cx, float x1, float y1, float x2, float y2, float x3, float y3 );
  void (* addArc)( GraphicsContext cx, float x, float y, float radius, float startAngle, float endAngle );
  void (* newPath)( GraphicsContext cx );
  void (* closePath)( GraphicsContext cx );

  void (* setLineWidth)( GraphicsContext cx, float w );

  void (* fill)( GraphicsContext cx );
  void (* stroke)( GraphicsContext cx );

  void (* save)( GraphicsContext cx );
  void (* restore)( GraphicsContext cx );

  void (* translate)( GraphicsContext cx, float x, float y );
  void (* scale)( GraphicsContext cx, float x, float y );

  void (* setFont)( GraphicsContext cx, int font, int size );
  void (* drawText)( GraphicsContext cx, float x, float y, const char* text, int alignment );
} GraphicsMethods;

extern GraphicsMethods gGraphicsMethods;

  #include "graphics.h"

extern GraphicsContext gcx;
#endif
#if ABC_INPUT_STRING
extern const char* abcInputString;
#endif

/*
 * macros and global variables
 */

/* -------------- general macros ------------- */

#define VERSION              "2.0"      /* version */
#define REVISION             ""         /* revison */
#define VDATE          "Jan 21 2007"     /* version date */
#define VERBOSE0           0            /* default verbosity */
#define DEBUG_LV           0            /* debug output level */
#define OUTPUTFILE      "Out.ps"        /* standard output file */
#define PLAY_TEMPO       120            /* default tempo for audio output */

/* default directory to search for format files */
#define DEFAULT_FDIR   ""

#define PAGEHEIGHT      29.7     /* basic page dimensions in cm ..  */
#define LEFTMARGIN       1.8     /* .. used in all default formats  */
#define STAFFWIDTH      17.4

#define DECO_IS_ROLL     0      /* ~ makes roll if 1, otherwise twiddle */

#define DEFVOICE        "1"     /* default name for first voice */


/* ----- macros for program internals ----- */

#define CM             28.35    /* factor to transform cm to pt */
#define PT              1.00    /* factor to transform pt to pt */
#define IN             72.00    /* factor to transform inch to pt */

#define STRL     1001           /* string length in info fields */
#define STRL1     255           /* string length for file names */
#define MAXSYMST   11           /* max symbols in start piece */
#define MAXHD      10           /* max heads on one stem */
#define NTEXT     100           /* max history lines for output */
#define MAXINF    100           /* max number of input files */
#define BSIZE    4001           /* buffer size for one input string */
#define BUFFSZ  50000           /* size of output buffer */
#define BUFFSZ1  3000           /* buffer reserved for one staff */
#define BUFFLN    100           /* max number of lines in output buffer */
#define NWPOOL   4000           /* char pool for vocals */
#define NWLINE      5           /* max number of vocal lines per staff */

#define BASE           192      /* base for durations */
#define WHOLE          192      /* whole note */
#define HALF            96      /* half note */
#define QUARTER         48      /* quarter note */
#define EIGHTH          24      /* 1/8 note */
#define SIXTEENTH       12      /* 1/16 note */
#define THIRTYSECOND     6      /* 1/32 note */
#define SIXTYFOURTH      3      /* 1/64 note */

/// Types of line found in an ABC file
enum {
  kLineComment =          1,      /* types of lines scanned */
  kLineMusic =            2, //FIXME-not used anywhere
  kLineToBecontinued =    3,
  kLineEOF  =             4,
  kLineInfo =             5,
  kLineTitle =            6,
  kLineMeter =            7,
  kLineParts =            8,
  kLineKey =              9,
  kLineXRef =            10,
  kLineDefaultNoteLength = 11,
  kLineHistory =         12,
  kLineBlank =           13,
  kLineWords =           14,  //FIXME-never seems to be tested for
  kLineMWords =          15,
  kLinePseudoComment =   16,
  kLineTempo =           17,
  kLineVoice =           18,
};

/// Types of symbol and lexer tokens
enum SymbolType {
  kSymbolInvisible = 1,
  kSymbolNote,
  kSymbolRest,
  kSymbolBar,
  kSymbolClef,
  kSymbolTimeSignature,
  kSymbolKeySignature,

  kSymbolSpace,
  kSymbolEscapeSequence,
  kSymbolContinue,
  kSymbolNewline,
  kSymbolDummy,
};

/// Types of bar line
enum {
  kBarSingle = 1,
  kBarDouble,           /* ||   thin double bar */
  kBarLeftRepeat,       /* |:   left repeat bar */
  kBarRightRepeat,      /* :|   right repeat bar */
  kBarDoubleRepeat,     /* ::   double repeat bar */
  kBarFat1,             /* [|   thick at section start */
  kBarFat2,             /* |]   thick at section end  */
  kBarInvisible         /* invisible; for endings without bars */
};

/// Types of accidental
enum {
  kAccidentalSharp = 1,
  kAccidentalNatural,
  kAccidentalFlat,
  kAccidentalDoubleSharp,
  kAccidentalDoubleFlat
};

// Types of decoration
enum {
  kDecorationGrace = 1,
  kDecorationStaccato,      // .
  kDecorationSlide,         // J - Not in the standard
  kDecorationEmBar,         // M - Not in the standard
  kDecorationHold,          // H - fermata
  kDecorationUpBow,         // u
  kDecorationDownBow,       // v
  kDecorationRoll,          // R
  kDecorationTrill,         // T
  kDecorationHat,           // K - Not in the standard
  kDecorationAtt,           // k - Not in the standard - what is it?

  // The following are in the standard but not supported
  kDecorationSegno,         // S
  kDecorationAccent,        // L
  kDecorationLowerMordent,  // M
  kDecorationCoda,          // O
  kDecorationUpperMordent,  // P

#if 0
  // From the 2.2 standard
  kDecorationTrill,           // "tr" (trill mark)
  kDecorationTrillStart,      // start of an extended trill
  kDecorationTrillEnd,        // end of an extended trill
  kDecorationLowerMordent,    // short /|/|/ squiggle with a vertical line through it
  kDecorationUpperMordent,    // short /|/|/ squiggle
  kDecorationMordent,         //     same as !lowermordent!
  kDecorationPralltriller,    // same as !uppermordent!
  kDecorationRoll,            // a roll mark (arc) as used in Irish music
  kDecorationTurn,            // a turn mark (also known as gruppetto)
  kDecorationTurnX,           // a turn mark with a line through it
  kDecorationInvertedTurn,    // an inverted turn mark
  kDecorationInvertedTurnX,   // an inverted turn mark with a line through it
  kDecorationArpeggio,        //    vertical squiggle
  kDecorationAccent,          // accent
  kDecorationFermata,         //     fermata or hold (arc above dot)
  kDecorationInvertedfermata, // upside down fermata
  kDecorationTenuto           // horizontal line to indicate holding note for full duration
  kDecoration0,               //     fingerings
  kDecoration1,               //     fingerings
  kDecoration2,               //     fingerings
  kDecoration3,               //     fingerings
  kDecoration4,               //     fingerings
  kDecoration5,               //     fingerings
  kDecoration+                //    left-hand pizzicato, or rasp for French horns
  kDecorationPlus,            // same as !+!
  kDecorationSnap,            // snap-pizzicato mark, visually similar to !thumb!
  kDecorationSlide,           // slide up to a note, visually similar to a half slur
  kDecorationWedge,           // small filled-in wedge mark
  kDecorationUpBow,           // V mark
  kDecorationDownBow,         //     squared n mark
  kDecorationOpen,            // small circle above note indicating open string or harmonic
  kDecorationThumb,           // cello thumb symbol
  kDecorationBreath,          // a breath mark (apostrophe-like) after note
  kDecorationP,               // dynamics marks
  kDecorationPP,              // dynamics marks
  kDecorationPPP,             // dynamics marks
  kDecorationPPPP,            // dynamics marks
  kDecorationMP,              // dynamics marks
  kDecorationMF,              // dynamics marks
  kDecorationF,               // dynamics marks
  kDecorationFF,              // dynamics marks
  kDecorationFFF,             // dynamics marks
  kDecorationFFFF,            // dynamics marks
  kDecorationSFZ,             // dynamics marks
  kDecorationCrescendoStart,  //  start of a < crescendo mark
  kDecorationCrescendoEnd,    //  end of a < crescendo mark, placed after the last note
  kDecorationDiminuendoStart, // start of a > diminuendo mark
  kDecorationDiminuendoEnd,   // end of a > diminuendo mark, placed after the last note
  kDecorationSegno,           // 2 ornate s-like symbols separated by a diagonal line
  kDecorationCoda,            // a ring with a cross in it
  kDecorationDS,              // the letters D.S. (=Da Segno)
  kDecorationDC,              // the letters D.C. (=either Da Coda or Da Capo)
  kDecorationDaCoda,          // the word "Da" followed by a Coda sign
  kDecorationDaCapo,          // the words "Da Capo"
  kDecorationFine,            // the word "fine"
  kDecorationShortPhrase,     // vertical line on the upper part of the staff
  kDecorationMediumPhrase,    // same, but extending down to the centre line
  kDecorationLongPhrase,      // same, but extending 3/4 of the way down

struct {
  const char* _name;
  Decorations _decoration;
} gDecorationNames[] = {
  { "trill", kDecorationTrill },
  { "trill(", kDecorationTrillStart },
  { "trill)", kDecorationTrillEnd },
  { "lowermordent", kDecorationLowerMordent },
  { "uppermordent", kDecorationUpperMordent },
  { "mordent", kDecorationMordent },
  { "pralltriller", kDecorationPralltriller },
  { "roll", kDecorationRoll },
  { "turn", kDecorationTurn },
  { "turnx", kDecorationTurnX },
  { "invertedturn", kDecorationInvertedTurn },
  { "invertedturnx", kDecorationInvertedTurnX },
  { "arpeggio", kDecorationArpeggio },
  { "accent", kDecorationAccent },
  { ">", kDecorationAccent} ,
  { "emphasis", kDecorationEmphasis },
  { "fermata", kDecorationFermata },
  { "invertedfermata", kDecorationInvertedfermata },
  { "tenuto", kDecorationTenuto },
  { "0", kDecoration0 },
  { "1", kDecoration1 },
  { "2", kDecoration2 },
  { "3", kDecoration3 },
  { "4", kDecoration4 },
  { "5", kDecoration5 },
  { "+", kDecorationPlus },
  { "plus", kDecorationPlus },
  { "snap", kDecorationSnap },
  { "slide", kDecorationSlide },
  { "wedge", kDecorationWedge },
  { "upbow", kDecorationUpBow },
  { "downbow", kDecorationDownBow },
  { "open", kDecorationOpen },
  { "thumb", kDecorationThumb },
  { "breath", kDecorationBreath },
  { "p", kDecorationP } ,
  { "pp", kDecorationPP } ,
  { "ppp", kDecorationPPP } ,
  { "pppp", kDecorationPPPP } ,
  { "mp", kDecorationMP } ,
  { "mf", kDecorationMF } ,
  { "f", kDecorationF } ,
  { "ff", kDecorationFF } ,
  { "fff", kDecorationFFF } ,
  { "ffff", kDecorationFFFF } ,
  { "sfz", kDecorationSFZ } ,
  { "crescendo(", kDecorationCrescendoStart },
  { "<(", kDecorationCrescendoStart },
  { "crescendo)", kDecorationCrescendoEnd },
  { "<)", kDecorationCrescendoEnd },
  { "diminuendo(", kDecorationDiminuendoStart },
  { ">(", kDecorationDiminuendoStart },
  { "diminuendo)", kDecorationDiminuendoEnd },
  { ">)", kDecorationDiminuendoEnd },
  { "segno", kDecorationSegno },
  { "coda", kDecorationCoda },
  { "D.S.", kDecorationDS },
  { "D.C.", kDecorationDC },
  { "dacoda", kDecorationDaCoda },
  { "dacapo", kDecorationDaCapo },
  { "fine", kDecorationFine },
  { "shortphrase", kDecorationShortPhrase },
  { "mediumphrase", kDecorationMediumPhrase },
  { "longphrase", kDecorationLongPhrase },
};
#endif
};

// Types of note head
enum {
  kHeadFull = 1,
  kHeadEmpty,
  kHeadOval,
};

/// Types of clef
enum {
  kClefTreble = 1,
  kClefTreble8,
  kClefBass,
  kClefAlto,
  kClefTenor,
};

enum {
  G_FILL =       1,            /* modes for glue */
  G_SHRINK =     2,
  G_SPACE =      3,
  G_STRETCH =    4,
};

enum {
  S_TITLE =      1,            /* where to do pattern matching */
  S_RHYTHM =     2,
  S_COMPOSER =   3,
  S_SOURCE =     4,
};

enum {
  TEXT_H =       1,            /* type of a text line */
  TEXT_W =       2,
  TEXT_Z =       3,
  TEXT_N =       4,
  TEXT_D =       5,
};

#define DO_OUTPUT    2            /* what program does */

#define STRLFMT    81             /* string length in FORMAT struct */


#define MAXNTEXT   400            /* for text output */
#define MAXWLEN     51

#define MAXTRP      10           /* max number of transpose flags */
#define MAXSVC      50           /* max number of voices for -S flag */


/* ----- global variables ------- */

extern int db;              /* debug control */

extern int maxSyms, maxVc;  /* for malloc */

#define NCOMP     5         /* max number of composer lines */

typedef struct              /* information fields */
{
  char area   [STRL];
  char book   [STRL];
  char comp   [NCOMP][STRL];
  int  ncomp;
  char disc   [STRL];
  char eskip  [STRL];
  char group  [STRL];
  char hist   [STRL];
  char info   [STRL];
  char key    [STRL];
  char len    [STRL];
  char meter  [STRL];
  char notes  [STRL];
  char orig   [STRL];
  char rhyth  [STRL];
  char src    [STRL];
  char title  [STRL];
  char title2 [STRL];
  char title3 [STRL];
  char parts  [STRL];
  char xref   [STRL];
  char trans  [STRL];
  char tempo  [STRL];
} InformationFields;

extern InformationFields info, default_info;

enum {
  kMaxGraceNotes = 20,
  kMaxDecorations = 10,
};

typedef struct                 /* describes grace notes */
{
  short _numNotes;              /* number of grace notes */
  short _pitch[kMaxGraceNotes]; /* pitches */
  short _accidental[kMaxGraceNotes]; /* accidentals */
} GraceNotes;

typedef struct                 /* describes decorations */
{
  short _numDecorations;       /* number of decorations */
  float _top;                  /* max height needed */
  unsigned char _type[kMaxDecorations]; /* type of deco */
} Decorations;

typedef struct Symbol          /* struct for a drawable symbol */
{
  unsigned char type;          /* type of symbol */
  unsigned char npitch;        /* number of note heads */
  unsigned char flags;         /* number of flags or bars */
  unsigned char dots;          /* number of dots */
  //JH: make into array of struct, not parallel
  unsigned char pits[MAXHD];   /* pitches for notes */
  unsigned char lens[MAXHD];   /* note lengths as multiple of BASE */
  unsigned char sl1 [MAXHD];   /* which slur start on this head */
  unsigned char sl2 [MAXHD];   /* which slur ends on this head */
  unsigned char accs[MAXHD];   /* code for accidentals */
  unsigned char ti1 [MAXHD];   /* flag to start tie here */
  unsigned char ti2 [MAXHD];   /* flag to end tie here */
  float shhd[MAXHD];           /* horizontal shift for heads */
  float shac[MAXHD];           /* horizontal shift for accidentals */
  short len;                   /* basic note length */
  short yadd;                  /* shift for treble/bass etc clefs */
  short ymn, ymx, yav;         /* min,mav,avg note head height */
  float x, y;                  /* position */
  float ylo, yhi;              /* bounds for this object */
  float xmn, xmx;              /* min,max h-pos of a head rel to top */
  unsigned char slur_st;       /* how many slurs starts here */
  unsigned char slur_end;      /* how many slurs ends here */
  char stem;                   /* 0,1,-1 for no stem, up, down */
  unsigned char head;          /* type of head */
  GraceNotes _graceNotes;      /* grace notes */
  Decorations _decorations;    /* decorations */
  float xs, ys;                /* position of stem end */
  short u, v, w, t, q;         /* auxillary information */
  float wl, wr;                /* left,right min width */
  float pl, pr;                /* left,right preferred width */
  float xl, xr;                /* left,right expanded width */
  unsigned char p_plet, q_plet, r_plet;  /* data for n-plets */
  float gchy;                  /* height of guitar chord */
  char text[41];               /* for guitar chords etc. */
  char* wordp[NWLINE];         /* pointers to wpool for vocals */
  int p;                       /* pointer to entry in posit table */
  float time;                  /* time for symbol start */
  unsigned fullmes : 1;        /* flag for full-measure rests */
  unsigned eoln : 1;           /* flag for last symbol in line */
  unsigned invis : 1;          /* mark note as invisible */
  unsigned word_st : 1;        /* 1 if word starts here */
  unsigned word_end : 1;       /* 1 if word ends here */
} Symbol;

void Symbol_print( Symbol* s );

typedef struct                 /* struct for a horizontal position */
{
  int type;                    /* type of symbols here */
  int next, prec;              /* pointers for linked list */
  int eoln;                    /* flag for line break */
  int* p;                      /* pointers to associated syms */
  float time, dur;             /* start time, duration */
  float wl, wr;                /* right and left widths */
  float space, shrink, stretch;  /* glue before this position */
  float tfac;                  /* factor to tune spacings */
  float x;                     /* final horizontal position */
} XPosition;


typedef struct          /* data to specify the meter */
{
  int meter1, meter2;
  int mflag, lflag;
  int dlen;
  int insert;
  char top[31];
} TimeSignature;

extern TimeSignature default_meter;

typedef struct            /* data to specify the key */
{
  int ktype;
  int sf;
  int add_pitch;
  int root, root_acc;
  int add_transp, add_acc[7];
} KeySignature;

extern KeySignature default_key;


typedef struct            /* struct to characterize a voice */
{
  char _id[33];                   /* identifier string, eg. a number */
  char name[81];                  /* full name of this voice */
  char sname[81];                 /* short name */
  char transpose[21];             /* transpose string */
  int halftones;                  /* number of halftones to transpose by */
  TimeSignature meter, meter0, meter1;    /* meter */
  KeySignature key, key0, key1;   /* keysig */
  int stems;                      /* +1 or -1 to force stem direction */
  int staves, brace, bracket;     /* for deco over several voices */
  int do_gch;                     /* 1 to output gchords for this voice */
  float sep;                      /* for space to next voice below */
  int nsym;                       /* number of symbols */
  int draw;                       /* flag if want to draw this voice */
  int select;                     /* flag if selected for output */
  int insert_btype, insert_num;   /* to split bars over linebreaks */
  int insert_bnum;                /* same for bar number */
  float insert_space;             /* space to insert after init syms */
  int end_slur;                   /* for a-b slurs */
  int rem_slur;                   /* for ties/slurs into boxes */
  char insert_text[81];           /* string over inserted barline */
} Voice;

/* --- things to alloc: */
extern Symbol* sym;               /* symbol list */
extern Symbol** symv;             /* symbols for voices */
extern XPosition* xp;             /* shared horizontal positions */
extern Voice* voice;              /* characteristics of a voice */
extern Symbol** sym_st;           /* symbols a staff start */
extern int* nsym_st;

extern Symbol zsym;               /* symbol containing zeros */

extern char lvoiceid[233];        /* string from last V: line */
extern int  nvoice, mvoice;       /* number of voices defd, nonempty */
extern int  ivc;                  /* current voice */
extern int  ivc0;                 /* top nonempty voice */
extern int ixpfree;               /* first free element in xp array */

typedef struct
{
  char _name[STRLFMT];
  float _size;
  int _box;
} FontSpec;

typedef struct Format                 /* --- struct for page layout */
{
  char     name[STRLFMT];
  float    pageheight, staffwidth;
  float    topmargin, botmargin, leftmargin;
  float    topspace, wordsspace, titlespace, subtitlespace, partsspace;
  float    composerspace, musicspace, vocalspace, textspace;
  float    scale, maxshrink, lineskipfac, parskipfac, indent;
  float    staffsep, sysstaffsep, systemsep;
  float    strict1, strict2;
  int      landscape, titleleft, continueall, breakall, writehistory;
  int      stretchstaff, stretchlast, withxrefs, barsperstaff;
  int      oneperpage, titlecaps, barnums;
  FontSpec titlefont, subtitlefont, vocalfont, textfont, tempofont;
  FontSpec composerfont, partsfont, gchordfont, wordsfont, voicefont;
  FontSpec barnumfont, barlabelfont, indexfont;
  float playtempo;
  char  auquality[21];
  char  accomp;
  int   barnums_cl;
} Format;

extern Format sfmt;                    /* format after initialization */
extern Format dfmt;                    /* format at start of tune */
extern Format cfmt;                    /* current format for output */

#if 0 //FIXME
typedef struct
{
  float _scale;
} ABCTypesetOptions;

typedef struct
{
  bool _trebleAccompaniment;
  bool _bassAccompaniment;
} ABCMIDIOptions;

typedef struct
{
  // Flags etc.
  ABCTypesetOptions _typesetOptions;
  ABCTextOutputOptions _textOutputOptions;
  ABCMIDIOptions _midiOptions;
  TransposeSpec _transpose;
  TextStream _inputStream;
  TextStream _outputStream;
} ABCSettings;

typedef struct
{
  // Symbols etc.
  int _tuneHeight;
} ABCData;

enum {
  kABCInformation,
  kABCWarning,
  kABCError,
};

typedef struct
{
  int _severity;
  const char* _message;
} ABCError;

extern ABCSettings abcSettings;

ABCProcess( ABCSettings * settings, ABCInputStream * stream, ABCData * data );
ABCWriteMIDI( ABCSettings * settings, ABCOutputStream * stream, ABCData * data );
ABCWriteGraphics( ABCSettings * settings, ABCOutputStream * stream, ABCData * data );
#endif

// Settings passed from outside, (command-line parameters)
typedef struct
{
  int landscape;                 /* flag for landscape output */
  float scalefac;                /* scale factor for symbol size */
  float lmargin;                 /* left margin */
  float swidth;                  /* staff width */
  float staffsep, dstaffsep;      /* staff separation */
  int bars_per_line;             /* bars for auto linebreaking */
  int continue_lines;            /* flag to continue all lines */
  int barnums;                   /* interval for bar numbers */
  int barnums_cl;                /* classical barnum counting ? */
  int write_history;             /* write history and notes ? */
  float playtempo;               /* tempo for audio output */
  float alfa_c;                  /* max compression allowed */
  float strict1, strict2;        /* 1stave, mstave strictness */
  int include_xrefs;             /* to include xref numbers in title */
  int break_continues;           /* ignore continuations ? */
  int make_index;                /* write index file ? */
  int make_audio;                /* write audio playback file ? */
  int midi_format;               /* MIDI file ? */
  int make_ps;                   /* write audio postscript file ? */
  int one_per_page;              /* new page for each tune ? */
  int do_mode;                   /* control whether to do index or output */
  char au_qstr[21];              /* code for au quality selection */
  char acc_pat[21];              /* accomp switch, pattern */
  int do_acc;                    /* generate accompaniement ? */
  int au_nomusic;                /* playback: only gchord acc */
  int pb_start, pb_end;          /* playback: start and end bars */
  int write_text;                /* write abc text file */
  int write_music;               /* write typeset sheet music */
  char transpose[MAXTRP][21];    /* target keys for transposition */
  char global_transpose[21];     /* global transposition */
  int global_halftones;          /* global transposition */
  int ntrp;                      /* number of tranpose specs */
  char vcselstr[101];            /* string for voice selection */
  int force_ktype;               /* for flag -S, global setting */
  int force_add_pitch;
  int frc_ktype[MAXSVC];         /* for flag -S, voice setting */
  int frc_add_pitch[MAXSVC];
  int pagenumbers;               /* write page numbers ? */
  int invert_chords;             /* invert chords in accompaniment ? */
  int select_all;                /* select all tunes ? */
  int respect_case;              /* does case matter in matching ? */
  int search_field0;             /* default search field */
  int pretty;                    /* for pretty but sprawling layout */
  int gmode;                     /* switch for glue treatment */

  /// For Cairo output:
#if ABC_USE_CAIRO
  char* outputFormat;   /// Image file format (e.g. "png")
#endif
} ABCMusicSettings;

extern ABCMusicSettings musicSettings;

//typedef struct {
//} ABCParser;
//
//extern ABCParser parser;

extern char fontnames[50][STRLFMT];           /* list of needed fonts */
extern int  nfontnames;

extern char buf[BUFFSZ];          /* output buffer.. should hold one tune */
extern int nbuf;                  /* number of bytes buffered */
extern float bposy;               /* current position in buffered data */
extern int   use_buffer;          /* 1 if lines are being accumulated */

extern char text [NTEXT][STRL];   /* pool for history, words, etc. lines */
extern int text_type[NTEXT];      /* type of each text line */
extern int ntext;                 /* number of text lines */
extern int linenum;               /* current line number in input file */
extern int tunenum;               /* number of current tune */
extern int numtitle;              /* how many titles were read */
extern int mline;                 /* number music lines in current tune */
extern int nsym;                  /* number of symbols in line */
extern int nsym0;                 /* nsym at start of parsing a line */
extern int pagenum;               /* current page in output file */
extern int writenum;              /* calls to write_buffer for each one tune */
extern int xrefnum;               /* xref number of current tune */
extern int do_meter, do_indent;   /* how to start next block */

extern int bagpipe;                   /* switch for HP mode */
extern int within_block; /* where we are in the file */
extern int do_this_tune;              /* are we typesetting the current one ? */
extern float posx, posy;               /* overall scale, position on page */
extern int barinit;                   /* carryover bar number between parts */

extern int num_ending;              /* number of endings to draw */
extern int mes1, mes2;               /* to count measures in an ending */

extern int overfull;                /* flag if staff overfull */
extern int do_words;                /* flag if staff has words under it */

extern int vb, verbose;             /* verbosity, global and within tune */
extern int in_page;

/* --- switches modified by flags: */
extern char znote[201];               /* annotation at top of page */

extern float alfa_last, beta_last;     /* for last short short line.. */

extern char outf[STRL1];              /* output file name */
extern char styf[STRL1];              /* layout style file name */
extern char styd[STRL1];              /* layout style directory */

extern int  file_open;                /* for output file */
extern int  file_initialized;         /* for output file */
extern FILE* fout, * ftext;     /* for output file */

#if !ABC_LIBRARY
extern char sel_str[MAXINF][STRL1];   /* list of selector strings */
extern int  s_field[MAXINF];          /* type of selection for each file */
extern int  psel[MAXINF];             /* pointers from files to selectors */
#endif

extern char gTuneParts[STRL];         // FIXME-part of parsed tune data
extern char gCurrentPart;             // FIXME-part of parsed tune data

extern char lout[BSIZE];

extern int utf8charset;
#endif

extern int ABCTransformABC( const char* abc, const char* midiFile, const char* spec );
