#include <string.h>     /* Included for strncmp */
#include <stdlib.h>     /* Included for malloc and free */
#include <stdio.h>

#include "abc2ps.h"
#include "parse.h"
#include "midi.h"

#define kMIDIFileHeader 0x4d546864
#define kMIDITrackHeader 0x4d54726b

#if ABC_PLAYBACK_PARTS
typedef struct {
  char* _data;
  int _length;
  int _allocated;
} String;

void String_new( String* s )
{
  s->_data = 0;
  s->_length = 0;
  s->_allocated = 0;
}

void String_free( String* s )
{
  if ( s->_data != 0 )
    free(s->_data);
  s->_data = 0;
  s->_length = 0;
  s->_allocated = 0;
}

int String_length( String* s )
{
  return s->_length;
}

char* String_data( String* s )
{
  return s->_data;
}

void String_reserve( String* s, int length )
{
  if ( length > s->_allocated ) {
    if ( s->_allocated == 0 )
      s->_allocated = 1;
    while ( length > s->_allocated )
      s->_allocated *= 2;
    s->_allocated = length;
    s->_data = (char*)realloc(s->_data, s->_allocated);
  }
}

void String_append( String* s, char c )
{
  String_reserve( s, s->_length+1 );
  s->_data[s->_length++] = c;
}

void String_appendString( String* s, String* s2 )
{
  String_reserve( s, s->_length+s2->_length );
  memcpy( s->_data+s->_length, s2->_data, s2->_length );
  s->_length += s2->_length;
}

void String_write( String* s, char* data, int length )
{
  String_reserve( s, s->_length+length );
  memcpy( s->_data+s->_length, data, length );
  s->_length += length;
}
#endif

#if ABC_PLAYBACK_PARTS
enum {
  kMaxParts = 26
};

static String sMIDIPartData[kMaxParts];
static String sMIDITuneData;
static String* sMIDIData;
#else
static long sTrackChunkPos;
#endif
static FILE* sMIDIFile;
static float sTimescale;
static float sStartTime;

static void write8bit(char data)
{
#if ABC_PLAYBACK_PARTS
  String_append( sMIDIData, data );
#else
  fputc(data, sMIDIFile);
#endif
}

static void writeVariableLength(long value)
{
  long buffer = value & 0x7f;

  while ((value >>= 7) > 0) {
    buffer <<= 8;
    buffer |= 0x80;
    buffer += (value & 0x7f);
  }
  while (1) {
    write8bit((char)(buffer & 0xff));

    if (buffer & 0x80)
      buffer >>= 8;
    else
      return;
  }
}

static void fwrite8bit(char data, FILE* f)
{
  fputc(data, sMIDIFile);
}

static void fwrite16bit(int data, FILE* f)
{
  fwrite8bit((char)((data & 0xff00) >> 8), f);
  fwrite8bit((char)(data & 0xff), f);
}

static void fwrite32bit(long data, FILE* f)
{
  fwrite8bit((char)((data >> 24) & 0xff), f);
  fwrite8bit((char)((data >> 16) & 0xff), f);
  fwrite8bit((char)((data >> 8 ) & 0xff), f);
  fwrite8bit((char)(data & 0xff), f);
}

struct Event
{
  char _type;
  char _pitch;
  int _time;
};

#define kMaxQueuedEvents 4096

static struct Event sEvents[kMaxQueuedEvents];
static int sNumEvents;
static int sLastEventTime;
static int sLastEventEmittedTime;
static int sTime; // Absolute time of last event
static int sLastPendingNoteOff;

static int eventCompare(const void* a, const void* b)
{
  struct Event* ea = (struct Event*)a;
  struct Event* eb = (struct Event*)b;
  if ( ea->_time == eb->_time ) {
    // Note off has to come before note on
    return (ea->_type & 0xf0) - (eb->_type & 0xf0);
  }
  if ( ea->_time < eb->_time )
    return -1;
  return 1;
}

static void flushEvents()
{
  //  printf("flush %d %d\n", sTime, sLastEventTime);
  qsort(sEvents, sNumEvents, sizeof(struct Event), eventCompare);
  for ( int i = 0; i < sNumEvents; i++ ) {
    struct Event* e = &sEvents[i];
    int deltaTime = e->_time - sLastEventEmittedTime;
    sLastEventEmittedTime = e->_time;
    //    printf(" -  emit %s %d %d %d\n", (e->_type & 0xf0) == 0x80 ? "off" : " on", e->_time/32, e->_type & 0x0f, e->_pitch);
    writeVariableLength(deltaTime);
    write8bit(e->_type);
    write8bit(e->_pitch);
    write8bit((e->_type & 0x0f) < 14 ? 0x60 : 0x50); // Melody is louder than acompaniment
  }

  sLastEventTime = 0;
  sNumEvents = 0;
  sLastPendingNoteOff = 0;
}

static void addEvent(int type, int pitch, int time)
{
  if ( sNumEvents >= kMaxQueuedEvents )
    flushEvents();
  struct Event* e = &sEvents[sNumEvents++];
  e->_type = type;
  e->_pitch = pitch;
  e->_time = time;
  //  printf(" + %s %d %d\n", e->_type == (char)0x80 ? "off" : " on", e->_time, e->_pitch);
}

static int readnump(const char** p)
{
  int n = 0;
  while ( '0' <= **p && **p <= '9' ) {
    n = n * 10 + **p - '0';
    (*p)++;
  }
  return n;
}

static void expandParts( const char* spec, char* out, int maxLen )
{
  char* outP = out;
  char* outPEnd = out+maxLen-2;
  const char* in;
  int i, j;
  int sp;
  char* stack[10];
  char lastch;

  sp = 0;
  in = spec;
  while (((*in >= 'A') && (*in <= 'Z')) || (*in == '(') || (*in == '.') ||
         (*in == ')') || (*in == '+') || (*in == '-') || 
         ((*in >= '0') && (*in <= '9'))) {
    if (*in == '.') {
      in = in + 1;
    };

    if ((*in >= 'A') && (*in <= 'Z')) {
      if ( outP < outPEnd )
        *outP++ = *in;
      lastch = *in;
      in++;
    };
    if (*in == '(') {
      if (sp < 10) {
        stack[sp] = outP;
        sp++;
      }
      else {
        syntax("nesting too deep in part specification", in);
      };
      in++;
    };
    if (*in == ')') {
      in = in + 1;
      if (sp > 0) {
        int repeats;
        char* start;
        char* stop;

        if ((*in >= '0') && (*in <= '9')) {
          repeats = readnump(&in);
        } else {
          repeats = 1;
        };
        sp--;
        start = stack[sp];
        stop = outP;
        for (i=1; i<repeats; i++) {
          for (j=0; j<((int) (stop-start)); j++) {
            if ( outP < outPEnd )
              *outP++ = *(start+j);
          };
        };
      }
      else {
        syntax("Too many )'s in part specification", in);
      };
    }
    else if ((*in >= '0') && (*in <= '9')) {
      int repeats;

      repeats = readnump(&in);
      if (outP != out) {
        for ( i = 1; i < repeats; i++ ) {
          if ( outP < outPEnd )
            *outP++ = lastch;
        };
      } else {
        syntax("No part to repeat in part specification", in);
      };
    };
  };
  if (sp != 0) {
    syntax("Too many ('s in part specification", in);
  };
  *outP++ = 0;
}


void midi_begin_tunes(FILE* fpin)
{
  sMIDIFile = fpin;
  sTime = 0;
  sLastEventTime = 0;
  sTimescale = 256;
  sNumEvents = 0;
  sLastEventEmittedTime = 0;
  sStartTime = 0;

  String_new( &sMIDITuneData );

#if !ABC_PLAYBACK_PARTS
  // Write the MIDI file header
  write32bit(kMIDIFileHeader);
  write32bit(6);
  write16bit(0); //format
  write16bit(1); //ntracks
  write16bit(60); //division

  // Write the MIDI track header
  write32bit(kMIDITrackHeader);
  sTrackChunkPos = ftell(sMIDIFile);
  write32bit(0);
#endif
}

void midi_begin_tune()
{
}

void midi_end_tune()
{
  char expandedParts[1024];

  if ( gTuneParts[0] == 0 ) {
    // If no P: in the tune header, work out a suitable one
    int n = 0;
    for ( int i = 0; i < kMaxParts; i++ )
      if ( String_length( &sMIDIPartData[i] ) != 0 )
        expandedParts[n++] = 'A'+i;
    if ( n == 0 )
      expandedParts[n++] = 'A';
    expandedParts[n++] = 0;
  }
  else
    expandParts( gTuneParts, expandedParts, sizeof(expandedParts) );

  for ( char* part = expandedParts; *part != 0; part++ ) {
    if ( *part >= 'A' && *part <= 'Z' ) {
      int i = *part - 'A';
      if ( String_length( &sMIDIPartData[i] ) != 0 ) {
        String_appendString( &sMIDITuneData, &sMIDIPartData[i] );
      }
    }
  }
  for ( int i = 0; i < kMaxParts; i++ )
    String_free( &sMIDIPartData[i] );
}

void midi_end_tunes()
{
  flushEvents();

#if ABC_PLAYBACK_PARTS
  // Write the MIDI file header
  fwrite32bit(kMIDIFileHeader, sMIDIFile);
  fwrite32bit(6, sMIDIFile);
  fwrite16bit(0, sMIDIFile); //format
  fwrite16bit(1, sMIDIFile); //ntracks
  fwrite16bit(60, sMIDIFile); //division

  midi_end_tune();
  int length = String_length(&sMIDITuneData);
//  char expandedParts[1024];
//  if ( gTuneParts[0] == 0 ) {
//    // If no P: in the tune header, work out a suitable one
//    int n = 0;
//    for ( int i = 0; i < kMaxParts; i++ )
//      if ( String_length( &sMIDIPartData[i] ) != 0 )
//        expandedParts[n++] = 'A'+i;
//    if ( n == 0 )
//      expandedParts[n++] = 'A';
//    expandedParts[n++] = 0;
//  }
//  else
//    expandParts( gTuneParts, expandedParts, sizeof(expandedParts) );
//  for ( char* part = expandedParts; *part != 0; part++ ) {
//    if ( *part >= 'A' && *part <= 'Z' ) {
//      int i = *part - 'A';
//      if ( String_length( &sMIDIPartData[i] ) != 0 )
//        length += String_length( &sMIDIPartData[i] );
//    }
//  }
  length += 4; // End of track event

  fwrite32bit(kMIDITrackHeader, sMIDIFile);
  fwrite32bit(length, sMIDIFile);

//  for ( char* part = expandedParts; *part != 0; part++ ) {
//    if ( *part >= 'A' && *part <= 'Z' ) {
//      int i = *part - 'A';
//      if ( String_length( &sMIDIPartData[i] ) != 0 ) {
//        fwrite(String_data( &sMIDIPartData[i] ), String_length( &sMIDIPartData[i] ), 1, sMIDIFile);
//#if 0//Dump MIDI track data
//        printf("Part: ");
//        for ( int j = 0; j < String_length( &sMIDIPartData[i] ); j++ )
//          printf("%02x ", String_data( &sMIDIPartData[i] )[j] & 0xff );
//        printf("\n");
//#endif
//      }
//    }
//  }
//  for ( int i = 0; i < kMaxParts; i++ )
//    String_free( &sMIDIPartData[i] );

  if ( String_length( &sMIDITuneData ) != 0 )
    fwrite(String_data( &sMIDITuneData ), String_length( &sMIDITuneData ), 1, sMIDIFile);
  String_free( &sMIDITuneData );

  // End of track
  fwrite32bit(0x00ff2f00, sMIDIFile);
#else
  // End of track
  write32bit(0x00ff2f00);

  size_t length = ftell(sMIDIFile) - sTrackChunkPos - 4;
  fseek(sMIDIFile, sTrackChunkPos, SEEK_SET);
  write32bit(length);
#endif
}

void midi_begin_part( char part )
{
#if ABC_PLAYBACK_PARTS
//  printf("Begin part %c\n", part);
  if ( part < 'A' || part > 'Z' )
    part = 'A';

  sMIDIData = &sMIDIPartData[part-'A'];
  String_reserve( sMIDIData, 2048 );
#endif
}

void midi_end_part()
{
//  printf("End part\n");
}

void midi_write_note(float t0, float tn, float tx, int pitch, float amplitude, int channel)
{
  pitch += 60;

  int onTime = (int)((sStartTime + t0) * sTimescale);
  int offTime = (int)((sStartTime + tn) * sTimescale);

  if ( offTime > sLastPendingNoteOff )
    sLastPendingNoteOff = offTime;
  //  printf(" - queue %s %d %d %d\n", " on", onTime/32, channel, pitch);
  //  printf(" - queue %s %d %d %d\n", "off", offTime/32, channel, pitch);
  //  if ( onTime < sLastEventEmittedTime )
  //    printf("Error: Going backwards in time! sLastEventEmittedTime=%d\n", sLastEventEmittedTime/32);
  sTime = onTime;
  //  if ( onTime >= sLastEventTime )
  //    flushEvents();
  //  if ( offTime > sLastEventTime )
  //    sLastEventTime = offTime;
  addEvent(0x90 + channel, pitch, onTime);
  addEvent(0x80 + channel, pitch, offTime);
}

void midi_set_start_time( float time )
{
  sStartTime = time;
}

void midi_flush_to_time( float time )
{
  flushEvents();
}
